(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(*  This example is based on an example designed for OCamlP3l.         *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rdec2.ml,v 1.14 2014/01/21 09:21:18 fclement Exp $ *)

type rectangle = {
  x0 : float; y0 : float;
  lx : float; ly : float
};;

type rectangular_grid = {
  domain : rectangle;
  nx : int; ny :int
};;

type computational_rectangular_grid = {
  mesh : rectangular_grid;
  px : int; py : int
};;

type regular_decomposition = {
  comp_grid : computational_rectangular_grid;
  nsbdx : int; nsbdy : int
};;

type subdomain = regular_decomposition * int;;

let rectangle_of x0 y0 lx ly = {
  x0 = x0; y0 = y0;
  lx = lx; ly = ly
};;

let rectangular_grid_of x0 y0 lx ly nx ny =
  let r = rectangle_of x0 y0 lx ly in {
    domain = r;
    nx = nx; ny = ny
  }
;;

let computational_rectangular_grid_of x0 y0 lx ly nx ny px py =
  let g = rectangular_grid_of x0 y0 lx ly nx ny in
  { mesh = g; px = px; py = py; }
;;

let number_of_subdomains_of_regular_decomposition dec =
  dec.nsbdx, dec.nsbdy
;;

let print_rectangle oc r =
  Printf.fprintf oc "%f %f\n%f %f\n" r.x0 r.y0 r.lx r.ly
;;

let print_rectangular_grid oc g =
  Printf.fprintf oc "%a%d %d\n"
    print_rectangle g.domain g.nx g.ny
;;

let print_computational_rectangular_grid oc g =
  Printf.fprintf oc "%a%d %d\n"
    print_rectangular_grid g.mesh g.px g.py
;;

(* (* Unused. *)
let print_regular_decomposition oc dec =
  Printf.fprintf oc "%a%d %d\n"
    print_computational_rectangular_grid
    dec.comp_grid dec.nsbdx dec.nsbdy
;;
*)

let read_rectangle ib =
  Scanf.bscanf ib "x0 = %f; y0 = %f;\nlx = %f; ly = %f;\n" rectangle_of
;;

let read_rectangular_grid ib =
  let r = read_rectangle ib in
  Scanf.bscanf ib "nx = %d; ny = %d;\n"
    (fun nx ny ->
      { domain = r; nx = nx; ny = ny; })
;;

let read_computational_rectangular_grid ib =
  let g = read_rectangular_grid ib in
  Scanf.bscanf ib "px = %d; py = %d;\n"
    (fun px py ->
      { mesh = g; px = px; py = py; })
;;

let read_regular_decomposition ib =
  let g = read_computational_rectangular_grid ib in
  Scanf.bscanf ib "nsbdx = %d; nsbdy = %d;\n"
    (fun nsbdx nsbdy ->
      { comp_grid = g; nsbdx = nsbdx; nsbdy = nsbdy; })
;;

exception Wrong_Input of string;;

let check_regular_decomposition {
    comp_grid = {
      mesh = {
        domain = { lx = lx; ly = ly; _ };
        nx = nx; ny = ny;
      };
      px = px; py = py;
    };
    nsbdx = nsbdx; nsbdy = nsbdy;
  } =
  if lx <= 0. || ly <= 0. then
    raise (Wrong_Input "Nonpositive length!");
  if nx <= 0 || ny <= 0 then
    raise (Wrong_Input "Nonpositive number of intervals!");
  if px <= 0 || py <= 0 then
    raise (Wrong_Input "Nonpositive overlapping!");
  if px > nx / nsbdx || py > ny / nsbdy then
    raise (Wrong_Input "Overlapping is too large!");
  if nsbdx <= 0 || nsbdy <= 0 then
    raise (Wrong_Input "Nonpositive number of subdomains!");
  if nsbdx > nx || nsbdy > ny then
    raise (Wrong_Input "Too many subdomains!");
  if nx mod nsbdx <> 0 || ny mod nsbdy <> 0 then
    raise (Wrong_Input
             "Number of subdomains do not divide number of intervals!");;

let parse_rdec_config_file fname =
  let ib =
    try Scanf.Scanning.from_file fname
    with Sys_error s -> prerr_endline s; exit 6 in
  read_regular_decomposition ib
;;

type location =
  | NorthSide
  | SouthSide
  | EastSide
  | WestSide
;;
module OrderedLocation = struct type t = location let compare = compare end;;
module Position = Set.Make (OrderedLocation);;

exception Out_of_range of int;;

let two_indices_numbering nx ny i =
  if i < 0 || i > nx * ny - 1 then raise (Out_of_range i) else
  (i mod nx, i / nx)
;;

let one_index_numbering nx ny (i, j) =
  (* FIXME *)
  if i < 0 || i > nx || j < 0 || j > ny then raise (Out_of_range (-1)) else
  i + j * nx
;;

let rectangular_grid_of_subdomain (dec, i) =
  let {
    comp_grid = {
      mesh = {
        domain = { x0 = x0; y0 = y0; lx = lx; ly = ly; };
        nx = nx; ny = ny;
      }; _
    };
    nsbdx = nsbdx; nsbdy = nsbdy;
  } = dec in
  let ix, iy = two_indices_numbering nsbdx nsbdy i in
  let lxi = lx /. float_of_int nsbdx
  and lyi = ly /. float_of_int nsbdy in
  let x0i = x0 +. float_of_int ix *. lxi
  and y0i = y0 +. float_of_int iy *. lyi in
  rectangular_grid_of
    x0i y0i lxi lyi (nx / nsbdx) (ny / nsbdy)
;;

let position_of_subdomain ({ nsbdx = nsbdx; nsbdy = nsbdy; _ }, i) =
  let ix, iy = two_indices_numbering nsbdx nsbdy i in
  let locations_to_add =
    [ (ix = 0, WestSide)
    ; (ix = nsbdx - 1, EastSide)
    ; (iy = 0, SouthSide)
    ; (iy = nsbdy - 1, NorthSide)
    ] in
  let rec loop accu = function
    | [] -> accu
    | (b, e) :: l -> if b then loop (Position.add e accu) l else loop accu l
  in
  loop Position.empty locations_to_add
;;

let complement pos =
  let fullset =
    Position.add WestSide
      (Position.add EastSide
         (Position.add SouthSide
            (Position.add NorthSide
               Position.empty))) in
  Position.diff fullset pos
;;

let neighbors_of (dec, i) =
  let (ix, iy) = two_indices_numbering dec.nsbdx dec.nsbdy i in
  let boundaries = complement (position_of_subdomain (dec, i)) in
  let add_neighbor f l = function
    | NorthSide -> f (ix, iy + 1) :: l
    | SouthSide -> f (ix, iy - 1) :: l
    | EastSide -> f (ix + 1, iy) :: l
    | WestSide -> f (ix - 1, iy) :: l
  in
  let f = one_index_numbering dec.nsbdx dec.nsbdy in
  List.fold_left (add_neighbor f) [] (Position.elements boundaries)
;;

let neighbor_orientation_of_id (dec, i) j =
  let (ix, iy) = two_indices_numbering dec.nsbdx dec.nsbdy i in
  let (jx, jy) = two_indices_numbering dec.nsbdx dec.nsbdy j in
  if ix = jx && iy + 1 = jy then Some NorthSide else
  if ix = jx && iy - 1 = jy then Some SouthSide else
  if ix + 1 = jx && iy = jy then Some EastSide else
  if ix - 1 = jx && iy = jy then Some WestSide else
  None
;;

let west_extension_of_computational_rectangular_grid {
    mesh = {
      domain = { x0 = x0; y0 = y0; lx = lx; ly = ly; };
      nx = nx; ny = ny;
    };
    px = px; py = py;
  } =
  let dx = lx /. float_of_int nx in
  computational_rectangular_grid_of
    (x0 -. float_of_int px *. dx) y0 (lx +. float_of_int px *. dx) ly
    (nx + px) ny px py

and east_extension_of_computational_rectangular_grid {
    mesh = {
      domain = { x0 = x0; y0 = y0; lx = lx; ly = ly; };
      nx = nx; ny = ny;
    };
    px = px; py = py;
  } =
  let dx = lx /. float_of_int nx in
  computational_rectangular_grid_of
    x0 y0
    (lx +. float_of_int px *. dx) ly
    (nx + px) ny px py

and south_extension_of_computational_rectangular_grid {
    mesh = {
      domain = { x0 = x0; y0 = y0; lx = lx; ly = ly; };
      nx = nx; ny = ny;
    };
    px = px; py = py;
  } =
  let dy = ly /. float_of_int ny in
  computational_rectangular_grid_of
    x0 (y0 -. float_of_int py *. dy)
    lx (ly +. float_of_int py *. dy)
    nx (ny + py) px py

and north_extension_of_computational_rectangular_grid {
    mesh = {
      domain = { x0 = x0; y0 = y0; lx = lx; ly = ly; };
      nx = nx; ny = ny;
    };
    px = px; py = py;
  } =
  let dy = ly /. float_of_int ny in
  computational_rectangular_grid_of
    x0 y0
    lx (ly +. float_of_int py *. dy)
    nx (ny + py) px py
;;

let computational_rectangular_grid_of_subdomain subd =
  let pos = position_of_subdomain subd in
  let dec = fst subd in
  let g0 = {
    mesh = rectangular_grid_of_subdomain subd;
    px = dec.comp_grid.px;
    py = dec.comp_grid.py;
  } in
  let ext_of_grid g = function
    | WestSide -> west_extension_of_computational_rectangular_grid g
    | EastSide -> east_extension_of_computational_rectangular_grid g
    | SouthSide -> south_extension_of_computational_rectangular_grid g
    | NorthSide -> north_extension_of_computational_rectangular_grid g
  in
  let rec loop accu = function
    | [] -> accu
    | cp :: l -> loop (ext_of_grid accu cp) l
  in
  loop g0 (Position.elements (complement pos))
;;

type edge_values = float array;;

type boundary = {
  west_ev : edge_values; east_ev : edge_values;
  south_ev : edge_values; north_ev : edge_values;
};;

type boundary_size = { size_x : int; size_y : int; };;

type interface = boundary array;;

let size_of_boundary bd =
  let size_x = Array.length bd.east_ev in
  let size_y = Array.length bd.north_ev in
  assert (size_x = Array.length bd.west_ev);
  assert (size_y = Array.length bd.south_ev);
  { size_x = size_x; size_y = size_y; }
;;

let print_edge_values oc ev =
  let f v = Printf.fprintf oc "%f\n" v in
  Array.iter f ev
;;

let print_boundary oc bd =
  Printf.fprintf oc "%a%a%a%a"
    print_edge_values bd.west_ev
    print_edge_values bd.east_ev
    print_edge_values bd.south_ev
    print_edge_values bd.north_ev;;

(* (* Unused. *)
let print_size_of_boundary oc bd =
  let sz = size_of_boundary bd in
  Printf.fprintf oc "{size_x = %i; size_y = %i}\n"
    sz.size_x sz.size_y
;;
*)

let read_edge_values ib sz =
  let f _i = Scanf.bscanf ib "%f\n" (fun v -> v) in
  Array.init sz f
;;


let read_boundary ib { size_x = size_x; size_y = size_y; } =
  let w = read_edge_values ib size_x in
  let e = read_edge_values ib size_x in
  let s = read_edge_values ib size_y in
  let n = read_edge_values ib size_y in
  { west_ev = w; east_ev = e; south_ev = s; north_ev = n; }
;;

let make_zero_edge_values n = Array.make n 0.;;

let make_zero_boundary { size_x = size_x; size_y = size_y; } =
  let w = make_zero_edge_values size_x and e = make_zero_edge_values size_x
  and s = make_zero_edge_values size_y and n = make_zero_edge_values size_y in
  { west_ev = w; east_ev = e; south_ev = s; north_ev = n; }
;;

let make_zero_interface dec =
  let f i =
    let { mesh = { nx = nx; ny = ny; _ }; _ } =
      computational_rectangular_grid_of_subdomain (dec, i) in
    let size_x = ny + 1
    and size_y = nx + 1 in
    make_zero_boundary {size_x = size_x; size_y = size_y} in
  Array.init (dec.nsbdx * dec.nsbdy) f
;;

let proj_boundary sd b1 (id, b2) =
  let copy = Array.copy in
  let len = Array.length in
  match neighbor_orientation_of_id sd id with
  | Some NorthSide ->
    assert (len b1.north_ev = len b2.south_ev);
    { b1 with north_ev = copy b2.south_ev }
  | Some SouthSide ->
    assert (len b1.south_ev = len b2.north_ev);
    { b1 with south_ev = copy b2.north_ev }
  | Some EastSide ->
    assert (len b1.east_ev = len b2.west_ev);
    { b1 with east_ev = copy b2.west_ev }
  | Some WestSide ->
    assert (len b1.west_ev = len b2.east_ev);
    { b1 with west_ev = copy b2.east_ev }
  | None -> invalid_arg "proj_boundary"
;;

let normalize_boundary b =
  let { north_ev = n; south_ev = s; east_ev = e; west_ev = w } = b in
  let l = Array.length in
  let mean x y = (x +. y) *. 0.5 in
  let nw = mean n.(0) w.(l w - 1) in
  let ne = mean n.(l n - 1) e.(l e - 1) in
  let sw = mean s.(0) w.(0) in
  let se = mean s.(l s - 1) e.(0) in
  n.(0) <- nw; w.(l w - 1) <- nw;
  n.(l n - 1) <- ne; e.(l e - 1) <- ne;
  s.(0) <- sw; w.(0) <- sw;
  s.(l s - 1) <- se; e.(0) <- se;
  b
;;

let for_all_subdomains (nsbdx, nsbdy) to_do =
  let num = ref (nsbdx * nsbdy - nsbdx) in
  for _j = nsbdy - 1 downto 0 do
    for _i = 0 to nsbdx - 1 do
      to_do !num;
      num := !num + 1;
    done;
    num := !num - 2 * nsbdx;
  done
;;
