(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2012 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: pgen.ml,v 1.11 2012/07/26 14:39:12 fclement Exp $ *)

(* Core pure caml functions *)

type choice = int list
type flagged_choice = choice * bool

let id x = x
let flip f x y = f y x

(* (* Unused. *)
let rec last = function
  | a :: [] -> a
  | _ :: t -> last t
  | [] -> raise Not_found
*)

let init_choice k =
  let rec gen l acc =
    if l >= k then acc else gen (l + 1) (l :: acc) in
  gen 0 []
(** Donne la premiere combinaison de [k] elements dans un ensemble
    d'au moins [k] elements. *)

let next_choice n l =
  let rec next_choice_aux l k =
    match l with
    | [] -> []
    | h :: t ->
      let t' = next_choice_aux t h in
      if h + 1 = k then h :: t' else
        if t' = t then h + 1 :: init_choice (List.length t)
        else h :: t' in
  next_choice_aux l n
(** Retourne la combinaison dans un ensemble de [n] elements qui suit
    [l]. *)

let (init_lchoice, recover_lchoice) =
  let rec iterate n f x acc =
    if n = 0 then acc else
    iterate (n - 1) f (f x) (x :: acc)
  in
  ( (fun k n p -> iterate p (next_choice n) (init_choice k) [])
  , (fun state n p -> iterate p (next_choice n) state [])
  )
(** [init_lchoice] Lift [init_choice] to the list level.
    [recover_lchoice st n p] Starts a new list of [p] permutations
    in a set of [n] elements. *)

let next_lchoice n vl =
  let rec do_n n f x =
    if n = 0 then x else do_n (n - 1) f (f x)
  in
  let l = List.length vl in
  List.map (do_n l (next_choice n)) vl
(** Lift [next_lchoice] to the list level. *)

let stop_condition n vl = not (
  List.exists id (snd (List.split vl)) ||
  (let l = fst (List.hd vl) in next_choice n l = l))
(** This is the stop condition of the main skeleton, we stop either if
    one combination has given a prime or if we reached the end of the
    combination set of our prime pool. *)

module IntSet =
  Set.Make(struct type t = int  let compare = compare end)

let get_uniq choice_list =
  let s =
    List.fold_left
      (fun s l ->
        List.fold_left (flip IntSet.add) s l)
      IntSet.empty choice_list in
  IntSet.elements s

let to_vfun f x = Array.of_list (f (Array.to_list x))

let get_vuniq = to_vfun get_uniq
let init_vchoice k n p = Array.of_list (init_lchoice k n p)
let recover_vchoice st n p = Array.of_list (recover_lchoice st n p)
let next_vchoice n = to_vfun (next_lchoice n)
let stop_vcond n v = stop_condition n (Array.to_list v)
(** Lifting to the array level. *)
